' based on code from https://www.youtube.com/watch?v=oJvJZNyW_rw

clear all

' Side Scroller using Bitmaps (optimized)

' Level storage
'    .    Sky
'    #    Solid block
'    G    Ground block
'    B    Brick block
'    ?    Question block
'    o    Coin
'    M    Mushroom
'    V    Flower
'    *    Star

nLevelWidth=64
nLevelHeight=16
sLevel$=""
sLevel$=sLevel$+"................................................................"
sLevel$=sLevel$+"................................................................"
sLevel$=sLevel$+".......ooooo.............................................MV*...."
sLevel$=sLevel$+"........ooo.............................................#####..."
sLevel$=sLevel$+".......................########................................."
sLevel$=sLevel$+".....BB?BBBB?BB.......###..............#.#......................"
sLevel$=sLevel$+"....................###................#.#......................"
sLevel$=sLevel$+"...................####........................................."
sLevel$=sLevel$+"GGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGG.##############.....########"
sLevel$=sLevel$+"...................................#.#...............###........"
sLevel$=sLevel$+"........................############.#............###..........."
sLevel$=sLevel$+"........................#............#.........###.............."
sLevel$=sLevel$+"........................#.############......###................."
sLevel$=sLevel$+"........................#................###...................."
sLevel$=sLevel$+"........................#################......................."
sLevel$=sLevel$+"................................................................"

' Player properties
fPlayerPosX=1.0
fPlayerPosY=1.0
fPlayerVelX=0.0
fPlayerVelY=0.0
bPlayerOnGround=false

' Camera properties
fCameraPosX=0.0
fCameraPosY=0.0

' Load sprite assets
bitmap 1 load sheet "SMB.xml" strip ".png"
bitmap 2 load "Duck.png"

' Load audio assets
sound 1 load "Jump.wav"
sound 2 load "Coin.wav"

' Initialize screen (320x200)
background 0,128,255 ' Sky
mode 4
double buffer

' Utility Lambdas
function GetLevelTile(x,y)
  x=int(x)
  y=int(y)
  if x>=0 and x<nLevelWidth and y>=0 and y<nLevelHeight then
    GetLevelTile=asc(sLevel$,y*nLevelWidth+x+1)
  else
    GetLevelTile=32
  end if
end function

function SetLevelTile(x,y,c)
  x=int(x)
  y=int(y)
  if x>=0 and x<nLevelWidth and y>=0 and y<nLevelHeight then
    sLevel$=strdelete$(sLevel$,y*nLevelWidth+x+1,1)
    sLevel$=strinsert$(sLevel$,chr$(c),y*nLevelWidth+x+1)
  end if
end function

cls
ink 255,255,255
print "Use the Left and Right arrow keys to move, Space to jump, Esc to quit."

tick:
fElapsedTime=elapsedtime/1000
if fElapsedTime>0.1 then fElapsedTime=0.1

' Handle input
if keydown(vk_up) then fPlayerVelY=-6.0
if keydown(vk_down) then fPlayerVelY=6.0
if keydown(vk_left) then
  t=-15.0:if bPlayerOnGround then t=-25.0
  fPlayerVelX=fPlayerVelX+t*fElapsedTime
end if
if keydown(vk_right) then
  t=15.0:if bPlayerOnGround then t=25.0
  fPlayerVelX=fPlayerVelX+t*fElapsedTime
end if

if keypressed(vk_esc) then end
if keypressed(vk_space) then
  if fPlayerVelY=0 then
    sound 1 play
    fPlayerVelY=-12.0
  end if
end if

' Gravity
fPlayerVelY=fPlayerVelY+20*fElapsedTime

' Drag
if bPlayerOnGround then
  fPlayerVelX=fPlayerVelX+(-3.0*fPlayerVelX*fElapsedTime)
  if abs(fPlayerVelX)<0.01 then
    fPlayerVelX=0.0
  end if
end if

' Clamp velocities
fPlayerVelX=clamp(fPlayerVelX,-10.0,10.0)
fPlayerVelY=clamp(fPlayerVelY,-10.0,10.0)

' Calculate potential new position
fNewPlayerPosX=fPlayerPosX+fPlayerVelX*fElapsedTime:if fNewPlayerPosX<0 then fNewPlayerPosX=0
fNewPlayerPosY=fPlayerPosY+fPlayerVelY*fElapsedTime:if fNewPlayerPosY<0 then fNewPlayerPosY=0

' Check for pickups!
call GetLevelTile(fNewPlayerPosX+0.0,fNewPlayerPosY+0.0) to t
if t=asc("o") then
  sound 2 play
  call SetLevelTile(fNewPlayerPosX+0.0,fNewPlayerPosY+0.0,asc("."))
end if

call GetLevelTile(fNewPlayerPosX+0.0,fNewPlayerPosY+1.0) to t
if t=asc("o") then
  sound 2 play
  call SetLevelTile(fNewPlayerPosX+0.0,fNewPlayerPosY+1.0,asc("."))
end if

call GetLevelTile(fNewPlayerPosX+1.0,fNewPlayerPosY+0.0) to t
if t=asc("o") then
  sound 2 play
  call SetLevelTile(fNewPlayerPosX+1.0,fNewPlayerPosY+0.0,asc("."))
end if

call GetLevelTile(fNewPlayerPosX+1.0,fNewPlayerPosY+1.0) to t
if t=asc("o") then
  sound 2 play
  call SetLevelTile(fNewPlayerPosX+1.0,fNewPlayerPosY+1.0,asc("."))
end if

' Check for collision
if fPlayerVelX<0 then ' Moving left
  call GetLevelTile(fNewPlayerPosX+0.0,fPlayerPosY+0.0) to t
  call GetLevelTile(fNewPlayerPosX+0.0,fPlayerPosY+0.9) to t2
  if t<>asc(".") or t2<>asc(".") then
    fNewPlayerPosX=int(fNewPlayerPosX)+1
    fPlayerVelX=0
  end if
else if fPlayerVelX>0 then ' Moving right
  call GetLevelTile(fNewPlayerPosX+1.0,fPlayerPosY+0.0) to t
  call GetLevelTile(fNewPlayerPosX+1.0,fPlayerPosY+0.9) to t2
  if t<>asc(".") or t2<>asc(".") then
    fNewPlayerPosX=int(fNewPlayerPosX)
    fPlayerVelX=0
  end if
end if

bPlayerOnGround=false
if fPlayerVelY<0 then ' Moving up
  call GetLevelTile(fNewPlayerPosX+0.0,fNewPlayerPosY) to t 
  call GetLevelTile(fNewPlayerPosX+0.9,fNewPlayerPosY) to t2
  if t<>asc(".") or t2<>asc(".") then
    fNewPlayerPosY=int(fNewPlayerPosY)+1
    fPlayerVelY=0
  end if
else if fPlayerVelY>0 then ' Moving down
  call GetLevelTile(fNewPlayerPosX+0.0,fNewPlayerPosY+1.0) to t
  call GetLevelTile(fNewPlayerPosX+0.9,fNewPlayerPosY+1.0) to t2
  if t<>asc(".") or t2<>asc(".") then
    fNewPlayerPosY=int(fNewPlayerPosY)
    fPlayerVelY=0
    bPlayerOnGround=true ' Player has a solid surface underfoot
  end if
end if

' Apply new position
fPlayerPosX=fNewPlayerPosX
fPlayerPosY=fNewPlayerPosY

' Link camera to player position
fCameraPosX=fPlayerPosX
fCameraPosY=fPlayerPosY

' Draw Level
nTileWidth=16
nTileHeight=16
nVisibleTilesX=ceil(bitmapwidth()/nTileWidth)
nVisibleTilesY=ceil(bitmapheight()/nTileHeight)

' Calculate top-leftmost visible tile
fOffsetX=fCameraPosX-nVisibleTilesX/2.0
fOffsetY=fCameraPosY-nVisibleTilesY/2.0

' Clamp camera to game boundaries
if fOffsetX<0 then fOffsetX=0
if fOffsetY<0 then fOffsetY=0
if fOffsetX>nLevelWidth-nVisibleTilesX then fOffsetX=nLevelWidth-nVisibleTilesX
if fOffsetY>nLevelHeight-nVisibleTilesY then fOffsetY=nLevelHeight-nVisibleTilesY

' Get offsets for smooth movement
fTileOffsetX=(fOffsetX-int(fOffsetX))*nTileWidth
fTileOffsetY=(fOffsetY-int(fOffsetY))*nTileHeight

' Draw visible tile map
clear bitmap
ink 255,255,255
for x=-1 to nVisibleTilesX
  for y=-1 to nVisibleTilesY
    call GetLevelTile(x+fOffsetX,y+fOffsetY) to sTileID
    select case chr$(sTileID)
      case "." ' Sky
        ' ignore
      case "#" ' Solid block
        stamp bitmap 1,"Solid" to x*nTileWidth-fTileOffsetX,y*nTileHeight-fTileOffsetY
      case "G" ' Ground block
        stamp bitmap 1,"Ground" to x*nTileWidth-fTileOffsetX,y*nTileHeight-fTileOffsetY
      case "B" ' Brick block
        stamp bitmap 1,"Brick" to x*nTileWidth-fTileOffsetX,y*nTileHeight-fTileOffsetY
      case "?" ' Question block
        stamp bitmap 1,"Question" to x*nTileWidth-fTileOffsetX,y*nTileHeight-fTileOffsetY
      case "o" ' Coin
        stamp bitmap 1,"Coin" to x*nTileWidth-fTileOffsetX,y*nTileHeight-fTileOffsetY
      case "M" ' Mushroom
        stamp bitmap 1,"Mushroom" to x*nTileWidth-fTileOffsetX,y*nTileHeight-fTileOffsetY
      case "V" ' Flower
        stamp bitmap 1,"Flower" to x*nTileWidth-fTileOffsetX,y*nTileHeight-fTileOffsetY
      case "*" ' Star
        stamp bitmap 1,"Star" to x*nTileWidth-fTileOffsetX,y*nTileHeight-fTileOffsetY
      case else
        ink 0,0,0
        rect fill x*nTileWidth-fTileOffsetX,y*nTileHeight-fTileOffsetY,(x+1)*nTileWidth-fTileOffsetX-1,(y+1)*nTileHeight-fTileOffsetY-1
        ink 255,255,255
    end select
  next y
next x

' Draw Player
'rect fill (fPlayerPosX-fOffsetX)*nTileWidth,(fPlayerPosY-fOffsetY)*nTileHeight,(fPlayerPosX-fOffsetX+1.0)*nTileWidth-1,(fPlayerPosY-fOffsetY+1.0)*nTileHeight-1
stamp bitmap 2 to (fPlayerPosX-fOffsetX)*nTileWidth,(fPlayerPosY-fOffsetY)*nTileHeight

' Swap buffers
swap buffers

' Print stats
locate 1,textrows-2
ink 255,255,255
highlight 0,0,0
print " " using$("###.#", fPlayerPosX+1) ", " using$("###.#", fPlayerPosY+1) " "
highlight 0

pause 0:goto tick
