' based on code from https://www.youtube.com/watch?v=oJvJZNyW_rw

clear all

' Side Scroller using TileMaps

' Level storage
'    .    Sky
'    #    Solid block
'    G    Ground block
'    B    Brick block
'    ?    Question block
'    o    Coin
'    M    Mushroom
'    V    Flower
'    *    Star

nLevelWidth=64
nLevelHeight=16
sLevel$=""
sLevel$=sLevel$+"................................................................"
sLevel$=sLevel$+"................................................................"
sLevel$=sLevel$+".......ooooo.............................................MV*...."
sLevel$=sLevel$+"........ooo.............................................#####..."
sLevel$=sLevel$+".......................########................................."
sLevel$=sLevel$+".....BB?BBBB?BB.......###..............#.#......................"
sLevel$=sLevel$+"....................###................#.#......................"
sLevel$=sLevel$+"...................####........................................."
sLevel$=sLevel$+"GGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGG.##############.....########"
sLevel$=sLevel$+"...................................#.#...............###........"
sLevel$=sLevel$+"........................############.#............###..........."
sLevel$=sLevel$+"........................#............#.........###.............."
sLevel$=sLevel$+"........................#.############......###................."
sLevel$=sLevel$+"........................#................###...................."
sLevel$=sLevel$+"........................#################......................."
sLevel$=sLevel$+"................................................................"

' Insert spaces to allow for smooth scrolling
for i=len(sLevel$)-nLevelWidth+1 to 2 step -nLevelWidth
  sLevel$=strinsert$(sLevel$,"  ",i)
next i
nLevelWidth=nLevelWidth+2
sLevel$=strrepeat$(" ",nLevelWidth)+" "+sLevel$+" "+strrepeat$(" ",nLevelWidth)
nLevelHeight=nLevelHeight+2

' Player properties
fPlayerPosX=2.0
fPlayerPosY=2.0
fPlayerVelX=0.0
fPlayerVelY=0.0
bPlayerOnGround=false

' Camera properties
fCameraPosX=0.0
fCameraPosY=0.0

redraw=true
iOldOffsetX=0
iOldOffsetY=0

' Load sprite assets
bitmap 1 load "SMB.png"
bitmap 2 load "Duck.png"
sprite 2 use bitmap 2 show xysize 4,4 xystep 4,4

' Load audio assets
sound 1 load "Jump.wav"
sound 2 load "Coin.wav"

' Initialize screen
background 0,128,255 ' Sky
tileset 2 new 16,16 use bitmap 1 map ".#GB?oMV*" to 0 map 32 to 0
tilemap 1 new 22,15 use tileset 2 target
double buffer
sprite 1 use tilemap 1 show scale 4,4 move relative 0,16

' Utility Lambdas
function GetLevelTile(x,y)
  x=int(x)
  y=int(y)
  if x>=0 and x<nLevelWidth and y>=0 and y<nLevelHeight then
    GetLevelTile=asc(sLevel$,y*nLevelWidth+x+1)
  else
    GetLevelTile=32
  end if
end function

function SetLevelTile(x,y,c)
  x=int(x)
  y=int(y)
  if x>=0 and x<nLevelWidth and y>=0 and y<nLevelHeight then
    sLevel$=strdelete$(sLevel$,y*nLevelWidth+x+1,1)
    sLevel$=strinsert$(sLevel$,chr$(c),y*nLevelWidth+x+1)
  end if
end function

cls
ink 255,255,255
print "Use the Left and Right arrow keys to move, Space to jump, Esc to quit."

tick:
fElapsedTime=elapsedtime/1000
if fElapsedTime>0.1 then fElapsedTime=0.1

' Handle input
if keydown(vk_up) then fPlayerVelY=-6.0
if keydown(vk_down) then fPlayerVelY=6.0
if keydown(vk_left) then
  t=-15.0:if bPlayerOnGround then t=-25.0
  fPlayerVelX=fPlayerVelX+t*fElapsedTime
end if
if keydown(vk_right) then
  t=15.0:if bPlayerOnGround then t=25.0
  fPlayerVelX=fPlayerVelX+t*fElapsedTime
end if

if keypressed(vk_esc) then end
if keypressed(vk_space) then
  if fPlayerVelY=0 then
    sound 1 play
    fPlayerVelY=-12.0
  end if
end if

' Gravity
fPlayerVelY=fPlayerVelY+20*fElapsedTime

' Drag
if bPlayerOnGround then
  fPlayerVelX=fPlayerVelX+(-3.0*fPlayerVelX*fElapsedTime)
  if abs(fPlayerVelX)<0.01 then
    fPlayerVelX=0.0
  end if
end if

' Clamp velocities
fPlayerVelX=clamp(fPlayerVelX,-10.0,10.0)
fPlayerVelY=clamp(fPlayerVelY,-10.0,10.0)

' Calculate potential new position
fNewPlayerPosX=fPlayerPosX+fPlayerVelX*fElapsedTime:fNewPlayerPosX=clamp(fNewPlayerPosX,1,nLevelWidth-2)
fNewPlayerPosY=fPlayerPosY+fPlayerVelY*fElapsedTime:fNewPlayerPosY=clamp(fNewPlayerPosY,1,nLevelHeight-2)

' Check for pickups!
call GetLevelTile(fNewPlayerPosX+0.0,fNewPlayerPosY+0.0) to t
if t=asc("o") then
  sound 2 play
  call SetLevelTile(fNewPlayerPosX+0.0,fNewPlayerPosY+0.0,asc("."))
  put tile fNewPlayerPosX-int(fOffsetX)+0.0,fNewPlayerPosY-int(fOffsetY)+0.0,0
end if

call GetLevelTile(fNewPlayerPosX+0.0,fNewPlayerPosY+1.0) to t
if t=asc("o") then
  sound 2 play
  call SetLevelTile(fNewPlayerPosX+0.0,fNewPlayerPosY+1.0,asc("."))
  put tile fNewPlayerPosX-int(fOffsetX)+0.0,fNewPlayerPosY-int(fOffsetY)+1.0,0
end if

call GetLevelTile(fNewPlayerPosX+1.0,fNewPlayerPosY+0.0) to t
if t=asc("o") then
  sound 2 play
  call SetLevelTile(fNewPlayerPosX+1.0,fNewPlayerPosY+0.0,asc("."))
  put tile fNewPlayerPosX-int(fOffsetX)+1.0,fNewPlayerPosY-int(fOffsetY)+0.0,0
end if

call GetLevelTile(fNewPlayerPosX+1.0,fNewPlayerPosY+1.0) to t
if t=asc("o") then
  sound 2 play
  call SetLevelTile(fNewPlayerPosX+1.0,fNewPlayerPosY+1.0,asc("."))
  put tile fNewPlayerPosX-int(fOffsetX)+1.0,fNewPlayerPosY-int(fOffsetY)+1.0,0
end if

' Check for collision
if fPlayerVelX<0 then ' Moving left
  call GetLevelTile(fNewPlayerPosX+0.0,fPlayerPosY+0.0) to t
  call GetLevelTile(fNewPlayerPosX+0.0,fPlayerPosY+0.9) to t2
  if t<>asc(".") or t2<>asc(".") then
    fNewPlayerPosX=int(fNewPlayerPosX)+1
    fPlayerVelX=0
  end if
else if fPlayerVelX>0 then ' Moving right
  call GetLevelTile(fNewPlayerPosX+1.0,fPlayerPosY+0.0) to t
  call GetLevelTile(fNewPlayerPosX+1.0,fPlayerPosY+0.9) to t2
  if t<>asc(".") or t2<>asc(".") then
    fNewPlayerPosX=int(fNewPlayerPosX)
    fPlayerVelX=0
  end if
end if

bPlayerOnGround=false
if fPlayerVelY<0 then ' Moving up
  call GetLevelTile(fNewPlayerPosX+0.0,fNewPlayerPosY) to t 
  call GetLevelTile(fNewPlayerPosX+0.9,fNewPlayerPosY) to t2
  if t<>asc(".") or t2<>asc(".") then
    fNewPlayerPosY=int(fNewPlayerPosY)+1
    fPlayerVelY=0
  end if
else if fPlayerVelY>0 then ' Moving down
  call GetLevelTile(fNewPlayerPosX+0.0,fNewPlayerPosY+1.0) to t
  call GetLevelTile(fNewPlayerPosX+0.9,fNewPlayerPosY+1.0) to t2
  if t<>asc(".") or t2<>asc(".") then
    fNewPlayerPosY=int(fNewPlayerPosY)
    fPlayerVelY=0
    bPlayerOnGround=true ' Player has a solid surface underfoot
  end if
end if

' Apply new position
fPlayerPosX=fNewPlayerPosX
fPlayerPosY=fNewPlayerPosY

' Link camera to player position
fCameraPosX=fPlayerPosX
fCameraPosY=fPlayerPosY

' Draw Level
nTileWidth=16
nTileHeight=16
nVisibleTilesX=tilemapcolumns()
nVisibleTilesY=tilemaprows()

' Calculate top-leftmost visible tile
fOffsetX=fCameraPosX-nVisibleTilesX/2.0
fOffsetY=fCameraPosY-nVisibleTilesY/2.0

' Clamp camera to game boundaries
if fOffsetX<0 then fOffsetX=0
if fOffsetY<0 then fOffsetY=0
if fOffsetX>nLevelWidth-nVisibleTilesX then fOffsetX=nLevelWidth-nVisibleTilesX
if fOffsetY>nLevelHeight-nVisibleTilesY then fOffsetY=nLevelHeight-nVisibleTilesY

' Get offsets for smooth movement
fTileOffsetX=(fOffsetX-int(fOffsetX))*nTileWidth
fTileOffsetY=(fOffsetY-int(fOffsetY))*nTileHeight

' Draw visible tile map
if redraw then
  UpdateScreen(0,0,nVisibleTilesX-1,nVisibleTilesY-1)
  redraw=false
  iOldOffsetX=int(fOffsetX)
  iOldOffsetX=int(fOffsetX)
end if

' Scroll the tile map
if int(fOffsetY)>iOldOffsetY then
  count=int(fOffsetY)-iOldOffsetY
  scroll tilemap up count
  UpdateScreen(0,nVisibleTilesY-count-1,nVisibleTilesX-1,nVisibleTilesY-1)
  iOldOffsetY=int(fOffsetY)
else if int(fOffsetY)<iOldOffsetY then
  count=iOldOffsetY-int(fOffsetY)
  scroll tilemap down count
  UpdateScreen(0,0,nVisibleTilesX-1,count-1)
  iOldOffsetY=int(fOffsetY)
end if  

if int(fOffsetX)>iOldOffsetX then
  count=int(fOffsetX)-iOldOffsetX
  scroll tilemap left count
  UpdateScreen(nVisibleTilesX-count-1,0,nVisibleTilesX-1,nVisibleTilesY-1)
  iOldOffsetX=int(fOffsetX)
else if int(fOffsetX)<iOldOffsetX then
  count=iOldOffsetX-int(fOffsetX)
  scroll tilemap right count
  UpdateScreen(0,0,count-1,nVisibleTilesY-1)
  iOldOffsetX=int(fOffsetX)
end if

' Swap buffers
shift tilemap fTileOffsetX,fTileOffsetY
swap buffers

' Move player
sprite 2 move (fPlayerPosX-fOffsetX)*nTileWidth-8,(fPlayerPosY-fOffsetY)*nTileHeight-8

' Print stats
locate 1,textrows-2
ink 255,255,255
highlight 0,0,0
print " " using$("###.#", fPlayerPosX) ", " using$("###.#", fPlayerPosY) " "
highlight 0

pause 0:goto tick

function UpdateScreen(x1,y1,x2,y2)
  for x=x1 to x2
    for y=y1 to y2
      call GetLevelTile(x+fOffsetX,y+fOffsetY) to sTileID
      put tile x,y,char(sTileID)
    next y
  next x
end function
