' based on code from https://www.youtube.com/watch?v=8OK8_tHeCIA

clear all
randomize timer

dim tetromino$(7)
nFieldWidth = 12
nFieldHeight = 18

nScreenWidth = 40  ' TileMap size in columns
nScreenHeight = 25 ' TileMap size in rows
xOffset=int((nScreenWidth-nFieldWidth)/2)
yOffset=int((nScreenHeight-nFieldHeight)/2)

function CustomCharBlock(ch,c)
  palette "X",c
  custom tile ch,0,"XXXXXXXX"
  custom tile ch,1,"X......X"
  custom tile ch,2,"X.XXXX.X"
  custom tile ch,3,"X.XXXX.X"
  custom tile ch,4,"X.XXXX.X"
  custom tile ch,5,"X.XXXX.X"
  custom tile ch,6,"X......X"
  custom tile ch,7,"XXXXXXXX"
end function

function CustomCharEqual(ch,c)
  palette "X",c
  custom tile ch,0,"........"
  custom tile ch,1,"........"
  custom tile ch,2,"XXXXXXXX"
  custom tile ch,3,"........"
  custom tile ch,4,"........"
  custom tile ch,5,"XXXXXXXX"
  custom tile ch,6,"........"
  custom tile ch,7,"........"
end function

function RotatePiece(px,py,r)
  select case r mod 4
    case 0
      RotatePiece=py*4+px
    case 1
      RotatePiece=12+py-(px*4)
    case 2
      RotatePiece=15-(py*4)-px
    case 3
      RotatePiece=3-py+(px*4)
  end select
end function

function DoesPieceFit(nTetromino,nRotation,nPosX,nPosY)
  ' All playfield cells > 0 are occupied
  for px=0 to 3
    for py=0 to 3
      ' Get index into piece
      call RotatePiece(px,py,nRotation) to idx
      ' Get index into field
      fi=(nPosY+py)*nFieldWidth+(nPosX+px)

      ' Check that test is in bounds. Note out of bounds does
      ' not necessarily mean a fail, as the long vertical piece
      ' can have cells that lie outside the boundary, so we'll
      ' just ignore them
      if nPosX+px>=0 and nPosX+px<nFieldWidth then
        if nPosY+py>=0 and nPosY+py<nFieldHeight then
          ' In bounds so do collision check
          if mid$(tetromino$(nTetromino),idx+1,1)<>"." and pField(fi)<>0 then
            DoesPieceFit=false
            exit function
          end if
        end if
      end if
    next py
  next px
  DoesPieceFit=true
end function

' Create TileMap Layer
tileset 2 copy tileset 1 target
tilemap 1 new nScreenWidth,nScreenHeight use tileset 2 target
double buffer
layer 1 use tilemap 1 show fit

' Create custom characters
palette ".",rgb(0,0,0)
CustomCharBlock(asc("A"),rgb(255,0,0))
CustomCharBlock(asc("B"),rgb(255,128,0))
CustomCharBlock(asc("C"),rgb(255,255,0))
CustomCharBlock(asc("D"),rgb(0,255,0))
CustomCharBlock(asc("E"),rgb(0,128,255))
CustomCharBlock(asc("F"),rgb(0,0,255))
CustomCharBlock(asc("G"),rgb(128,0,255))
CustomCharBlock(asc("#"),rgb(128,128,128))
CustomCharEqual(asc("="),RGB(255,255,255))

' Tetronimos 4x4
tetromino$(0)="..X...X...X...X."
tetromino$(1)="..X..XX...X....."
tetromino$(2)=".....XX..XX....."
tetromino$(3)="..X..XX..X......"
tetromino$(4)=".X...XX...X....."
tetromino$(5)=".X...X...XX....."
tetromino$(6)="..X...X..XX....."

' Playfield
dim pField(nFieldWidth*nFieldHeight)
for x=0 to nFieldWidth-1
  for y=0 to nFieldHeight-1
    if x=0 or x=nFieldWidth-1 or y=nFieldHeight-1 then
      pField(y*nFieldWidth+x)=9
    else
      pField(y*nFieldWidth+x)=0
    end if
  next y
next x

' Game logic
dim bKey(3)
nNextPiece=range(0,6)
nCurrentPiece=range(0,6)
nCurrentRotation=0
nCurrentX=nFieldWidth/2-2
nCurrentY=0
nSpeed=20
nSpeedCount=0
bForceDown=false
bRotateHold=true
nPieceCount=0
nScore=0
dim vLines(nFieldHeight):vp=0
bGameOver=false

while not bGameOver
  ' Timing =======================
  pause 50
  nSpeedCount=nSpeedCount+1
  bForceDown=(nSpeedCount=nSpeed)
  
  ' Input ========================
  ' L R D Z
  bKey(0)=keydown(vk_left)
  bKey(1)=keydown(vk_right)
  bKey(2)=keydown(vk_down)
  bKey(3)=keydown(asc("z"))
  
  if keypressed(vk_esc) then bGameOver=true:exit while
  
  ' Game Logic ===================

  ' Handle player movement
  if bKey(0) then call DoesPieceFit(nCurrentPiece,nCurrentRotation,nCurrentX-1,nCurrentY) to b:if b then nCurrentX=nCurrentX-1
  if bKey(1) then call DoesPieceFit(nCurrentPiece,nCurrentRotation,nCurrentX+1,nCurrentY) to b:if b then nCurrentX=nCurrentX+1
  if bKey(2) then call DoesPieceFit(nCurrentPiece,nCurrentRotation,nCurrentX,nCurrentY+1) to b:if b then nCurrentY=nCurrentY+1
  
  ' Rotate, but latch to stop wild spinning
  if bKey(3) then
    call DoesPieceFit(nCurrentPiece,nCurrentRotation+1,nCurrentX,nCurrentY) to b
    if bRotateHold and b then nCurrentRotation=nCurrentRotation+1
    bRotateHold=false
  else
    bRotateHold=true
  end if
  
  ' Force the piece down the playfield if it's time
  if bForceDown then
    ' Update difficulty every 50 pieces
    nSpeedCount=0
    nPieceCount=nPieceCount+1
    if nPieceCount mod 50=0 then if nSpeed>=10 then nSpeed=nSpeed-1
    
    ' Test if piece can be moved down
    call DoesPieceFit(nCurrentPiece,nCurrentRotation,nCurrentX,nCurrentY+1) to b
    if b then
      ' It can, so do it!
      nCurrentY=nCurrentY+1
    else
      ' It can't! Lock the piece in place
      for px=0 to 3
        for py=0 to 3
          call RotatePiece(px,py,nCurrentRotation) to r
          if mid$(tetromino$(nCurrentPiece),r+1,1)<>"." then pField((nCurrentY+py)*nFieldWidth+(nCurrentX+px))=nCurrentPiece+1
        next py
      next px
            
      ' Check for completed lines
      for py=0 to 3
        if nCurrentY+py<nFieldHeight-1 then
          bLine=true
          for px=1 to nFieldWidth-2
            bLine=bLine and (pField((nCurrentY+py)*nFieldWidth+px))<>0
          next px
          if bLine then
            ' Remove line, set to =
            for px=1 to nFieldWidth-2
              pField((nCurrentY+py)*nFieldWidth+px)=8
            next px
            vLines(vp)=nCurrentY+py
            vp=vp+1
          end if
        end if
      next py  
      
      nScore=nScore+25
      if vp>0 then nScore=nScore+(1 shl vp)*100

      ' Pick new piece
      nCurrentX=nFieldWidth/2-2
      nCurrentY=0
      nCurrentRotation=0
      nCurrentPiece=nNextPiece
      nNextPiece=range(0,6)

      ' If piece does not fit straight away, game over!
      call DoesPieceFit(nCurrentPiece,nCurrentRotation,nCurrentX,nCurrentY) to b
      bGameOver=not b
    end if
  end if
     
  ' Display ======================

  clear tilemap
  
  ' Draw playfield
  for x=0 to nFieldWidth-1
    for y=0 to nFieldHeight-1
      put tile x+xOffset,y+yOffset,asc(mid$(" ABCDEFG=#",pField(y*nFieldWidth+x)+1,1))
    next y
  next x

  ' Draw current piece
  for px=0 to 3
    for py=0 to 3
      call RotatePiece(px,py,nCurrentRotation) to r
      if mid$(tetromino$(nCurrentPiece),r+1,1)<>"." then 
        put tile nCurrentX+px+xOffset,nCurrentY+py+yOffset,nCurrentPiece+65
      end if
    next py
  next px
  
  ' Draw next piece
  for px=0 to 3
    for py=0 to 3
      call RotatePiece(px,py,0) to r
      if mid$(tetromino$(nNextPiece),r+1,1)<>"." then
        put tile px+xOffset+nFieldWidth+2,py+yOffset,nNextPiece+65
      else
        put tile px+xOffset+nFieldWidth+2,py+yOffset,32
      end if
    next py
  next px
  
  ' Swap buffers
  swap buffers

  ' Draw score
  locate 0,0
  print "Score: ";nScore

  ' Animate line completion
  if vp>0 then
    ' Display frame
    pause 400
    
    ' Remove the completed lines
    for i=0 to vp-1
      v=vLines(i)
      for px=1 to nFieldWidth-2
        for py=v to 1 step -1
          pField(py*nFieldWidth+px)=pField((py-1)*nFieldWidth+px)
        next py
        pField(px)=0
      next px
    next i
    vp=0
  end if
wend

print "Game Over!! Score: ";nScore
waitkey a$
end
