clear all

' medium.com/@dakk/quantum-computing-on-a-commodore-64-in-200-lines-of-basic-eda7658b32a4

rem r* are real parts and i* are imagenary parts of the sv
r0 = 1 : i0 = 0 : r1 = 0 : i1 = 0
r2 = 0 : i2 = 0 : r3 = 0 : i3 = 0
a0 = 0
shots = 28
randomize timer
cls
print "c64 quantum simulator"
print "created by davide gessa (dakk)"
print "enter gate seq (x0,x1,y0,y1,z0,z1,h0,h1,cx,sw)"
print "e.g. h0cx"
input g$
g$=lcase$(trim$(g$))
print "calculating the statevector...";
for i = 1 to len(g$) step 2
  gate$ = mid$(g$, i, 2)
  gosub simulate
  print ".";
next i
print

sq = r0*r0 + i0*i0 + r1*r1 + i1*i1 + r2*r2 + i2*i2 + r3*r3 + i3*i3
if abs(sq - 1) > 0.00001 then gosub normalize

print "running " shots " iterations..."
z0 = 0 : z1 = 0 : z2 = 0 : z3 = 0
p0 = (r0 * r0 + i0 * i0)
p1 = (r1 * r1 + i1 * i1) + p0
p2 = (r2 * r2 + i2 * i2) + p1
p3 = (r3 * r3 + i3 * i3) + p2

for i = 1 to shots
  r = rnd(1)
  if r < p0 then z0 = z0 + 1
  if r >= p0 and r < p1 then z1 = z1 + 1
  if r >= p1 and r < p2 then z2 = z2 + 1
  if r >= p2 and r < p3 then z3 = z3 + 1
next i

print "results:"
print "00: ["z0"] "; : for i = 1 to z0 : print "O"; : next i : print
print "01: ["z2"] "; : for i = 1 to z2 : print "O"; : next i : print 
print "10: ["z1"] "; : for i = 1 to z1 : print "O"; : next i : print
print "11: ["z3"] "; : for i = 1 to z3 : print "O"; : next i : print
goto done

rem simulate gate operation
simulate:
if gate$ = "x0" then gosub x0Gate
if gate$ = "x1" then gosub x1Gate
if gate$ = "y0" then gosub y0Gate
if gate$ = "y1" then gosub y1Gate
if gate$ = "z0" then gosub z0Gate
if gate$ = "z1" then gosub z1Gate
if gate$ = "h0" then gosub h0Gate
if gate$ = "h1" then gosub h1Gate
if gate$ = "cx" then gosub cxGate
if gate$ = "sw" then gosub swGate
return

rem x0 gate
x0Gate:
a0 = r0 : r0 = r1 : r1 = a0
a0 = i0 : i0 = i1 : i1 = a0
a0 = r2 : r2 = r3 : r3 = a0
a0 = i2 : i2 = i3 : i3 = a0
return

rem x1 gate
x1Gate:
a0 = r1 : r1 = r3 : r3 = a0
a0 = i1 : i1 = i3 : i3 = a0
a0 = r0 : r0 = r2 : r2 = a0
a0 = i0 : i0 = i2 : i2 = a0

rem y0 gate
y0Gate:
a0 = i0 : i0 = -r0 : r0 = a0
a0 = i1 : i1 = -r1 : r1 = a0
a0 = i2 : i2 = -r2 : r2 = a0
a0 = i3 : i3 = -r3 : r3 = a0
return

rem y1 gate
y1Gate:
a0 = i1 : i1 = -r1 : r1 = a0
a0 = i3 : i3 = -r3 : r3 = a0
return

rem z0 gate
z0Gate:
i2 = -i2 : i3 = -i3
return

rem z1 gate
z1Gate:
i1 = -i1 : i3 = -i3
return

rem h0 gate
h0Gate:
a0 = (r0 + r1) / sqr(2) : a1 = (i0 + i1) / sqr(2)
b0 = (r0 - r1) / sqr(2) : b1 = (i0 - i1) / sqr(2)
r0 = a0 : i0 = a1 : r1 = b0 : i1 = b1
a0 = (r2 + r3) / sqr(2) : a1 = (i2 + i3) / sqr(2)
b0 = (r2 - r3) / sqr(2) : b1 = (i2 - i3) / sqr(2)
r2 = a0 : i2 = a1 : r3 = b0 : i3 = b1
return

rem h1 gate
h1Gate:
a0 = (r0 + r2) / sqr(2) : a1 = (i0 + i2) / sqr(2)
b0 = (r0 - r2) / sqr(2) : b1 = (i0 - i2) / sqr(2)
r0 = a0 : i0 = a1 : r2 = b0 : i2 = b1
a0 = (r1 + r3) / sqr(2) : a1 = (i1 + i3) / sqr(2)
b0 = (r1 - r3) / sqr(2) : b1 = (i1 - i3) / sqr(2)
r1 = a0 : i1 = a1 : r3 = b0 : i3 = b1
return

rem cx gate
cxGate:
a0 = r1 : r1 = r3 : r3 = a0
a0 = i1 : i1 = i3 : i3 = a0
return 

rem sw gate
swGate:
a0 = r1 : r1 = r2 : r2 = a0
a0 = i1 : i1 = i2 : i2 = a0
return

rem statevcector normalization
normalize:
nf = sqr(1 / sq)
r0 = r0 * nf
i0 = i0 * nf
r1 = r1 * nf
i1 = i1 * nf
r2 = r2 * nf
i2 = i2 * nf
r3 = r3 * nf
i3 = i3 * nf
return

done:
waitkey a$
end
